package org.kantega.jexmec.spring;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.*;
import org.kantega.jexmec.simple.DefaultPluginManager;
import org.kantega.jexmec.simple.ServicesProxyFactory;
import org.kantega.jexmec.simple.StaticPluginClassLoaderProvider;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.config.AbstractFactoryBean;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ApplicationContextAware;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationListener;
import org.springframework.context.event.ContextRefreshedEvent;

import java.util.Collections;
import java.util.List;

/**
 * FactoryBean for easy creation and startup of a DefaultPluginManager in a Spring context.
 */
public class PluginManagerFactoryBean extends AbstractFactoryBean implements ApplicationContextAware, ApplicationListener {
    private ApplicationContext applicationContext;


    private Class pluginClass;

    private Class serviceClass;


    private ServiceLocator serviceLocator;

    private List<PluginLoader<Plugin>> pluginLoaders = Collections.emptyList();

    private List<PluginClassLoaderProvider> pluginClassLoaderProviders = Collections.emptyList();

    private boolean addSystemClassLoaderPluginLoader = true;


    public Class getObjectType() {
        return DefaultPluginManager.class;
    }

    public boolean isSingleton() {
        return true;
    }

    public DefaultPluginManager createInstance() throws Exception {

        final DefaultPluginManager manager =  new DefaultPluginManager(pluginClass);

        if (serviceLocator == null && serviceClass != null) {
            serviceLocator = new SpringServiceLocator(applicationContext, serviceClass);
        }

        if(serviceLocator != null) {
            manager.addServiceLocator(serviceLocator);
            if(serviceClass != null) {
                manager.addService(ServiceKey.by(serviceClass), ServicesProxyFactory.create(serviceClass, serviceLocator));
            }
        }

        for (PluginLoader<Plugin> loader : pluginLoaders) {
            manager.addPluginLoader(loader);
        }



        if(addSystemClassLoaderPluginLoader) {
            manager.addPluginClassLoaderProvider(new StaticPluginClassLoaderProvider(getClass().getClassLoader()));
        }

        for(PluginClassLoaderProvider provider : pluginClassLoaderProviders) {
            manager.addPluginClassLoaderProvider(provider);
        }

        return manager;
    }

    public void setApplicationContext(ApplicationContext applicationContext) throws BeansException {
        this.applicationContext = applicationContext;
    }


    public void setPluginLoaders(List<PluginLoader<Plugin>> pluginLoaders) {
        this.pluginLoaders = pluginLoaders;
    }

    public void setPluginClassLoaderProviders(List<PluginClassLoaderProvider> pluginClassLoaderProviders) {
        this.pluginClassLoaderProviders = pluginClassLoaderProviders;
    }

    public void setServiceLocator(ServiceLocator serviceLocator) {
        this.serviceLocator = serviceLocator;
    }

    public void onApplicationEvent(ApplicationEvent event) {
        if (event instanceof ContextRefreshedEvent && event.getSource() == applicationContext) {
            try {
                DefaultPluginManager manager = (DefaultPluginManager) getObject();

                manager.start();
            } catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
    }

    public void setServiceClass(Class serviceClass) {
        this.serviceClass = serviceClass;
    }

    public void setPluginClass(Class pluginClass) {
        this.pluginClass = pluginClass;
    }


    public void setAddSystemClassLoaderProvider(boolean addSystemClassLoaderPluginLoader) {
        this.addSystemClassLoaderPluginLoader = addSystemClassLoaderPluginLoader;
    }

    @Override
    public void afterPropertiesSet() throws Exception {
        super.afterPropertiesSet();
        if(pluginClass == null) {
            throw new IllegalArgumentException("You have not specified which pluginClass to use. Please set the pluginClass property to be your plugin class");
        }
    }

    @Override
    protected void destroyInstance(Object instance) throws Exception {
        DefaultPluginManager.class.cast(instance).stop();
    }
}
