package org.kantega.jexmec.spring;

/*
 * Copyright 2010 Kantega AS
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import org.kantega.jexmec.Plugin;
import org.kantega.jexmec.PluginLoader;
import org.kantega.jexmec.ServiceKey;
import org.kantega.jexmec.ServiceLocator;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.annotation.AutowiredAnnotationBeanPostProcessor;
import org.springframework.beans.factory.config.BeanFactoryPostProcessor;
import org.springframework.beans.factory.config.ConfigurableListableBeanFactory;
import org.springframework.beans.factory.support.BeanDefinitionRegistry;
import org.springframework.beans.factory.support.RootBeanDefinition;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.AbstractApplicationContext;
import org.springframework.context.support.ClassPathXmlApplicationContext;

import java.io.IOException;
import java.net.URL;
import java.util.*;

/**
 * Implementation of {@link PluginLoader} based on the Spring Framework
 * Plugins are located and loaded by creating a Spring Application Context read from the given or
 * default paths.
 *
 * The applicaton context is set up with annotation based autowiring and all services in
 * {@link org.kantega.jexmec.ServiceLocator} is added as autowirable services.
 * @param <P> the Plugin implementation type this instance will load
 */
public class SpringPluginLoader<P extends Plugin> implements PluginLoader<P> {

    private final Map<ClassLoader, Collection<ApplicationContext>> contexts = Collections.synchronizedMap(new HashMap<ClassLoader, Collection<ApplicationContext>>());
    private final String[] contextLocations;

    private List<BeanFactoryPostProcessor> beanFactoryPostProcessors;

    /**
     * Create a SpringPluginLoader looking for Spring contexts in the given locations.
     * @param contextLocations the locations to look for Spring Application Context files.
     */
    public SpringPluginLoader(String... contextLocations) {
        this.contextLocations = contextLocations;
    }

    /**
     * Create a SpringPluginLoader loading Spring contexts in the default locations:
     * <code>META-INF/services/MyPlugin/spring.xml</code> and <code>META-INF/services/com.mycompany.MyPlugin/spring.xml</code>.
     */
    public SpringPluginLoader() {
        this.contextLocations = null;
    }

    public void start() {
        
    }

    /**
     * Stop the SpringPluginLoader, destroying any ApplicationContexts
     */
    public void stop() {
        for(ClassLoader loader : contexts.keySet()) {
            destroyContexts(loader);
        }
        contexts.clear();
    }

    private void destroyContexts(ClassLoader loader) {
        final Collection<ApplicationContext> classLoaderContexts = contexts.get(loader);
        if(classLoaderContexts != null) {
            for(ApplicationContext context : classLoaderContexts) {
                if(context instanceof AbstractApplicationContext) {
                    AbstractApplicationContext ctx = (AbstractApplicationContext) context;
                    ctx.destroy();
                }
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public List<P> loadPlugins(Class<P> pluginClass, ClassLoader classLoader, ServiceLocator serviceLocator) {

        String[] contextLocations = getContextLocations(pluginClass);

        List<P> plugins = new ArrayList<P>();

        try {

            for(String contextLocation : contextLocations) {
                final List<URL> resources = Collections.list(classLoader.getResources(contextLocation));

                for(URL resource : resources) {
                    ApplicationContext context = initApplicationContext(serviceLocator, resource, classLoader);

                    Collection<ApplicationContext> classLoaderContexts = contexts.get(classLoader);
                    if(classLoaderContexts == null) {
                        contexts.put(classLoader,  classLoaderContexts = new HashSet<ApplicationContext>());
                    }
                    classLoaderContexts.add(context);
                    plugins.addAll(context.getBeansOfType(pluginClass).values());
                }

            }

            return plugins;

        } catch (IOException e) {
            throw new RuntimeException(e);
        }

    }

    /**
     * Destroy any Application contexts loaded from the given ClassLoader.
     * @param classLoader
     */
    public void unloadPlugins(ClassLoader classLoader) {
        destroyContexts(classLoader);
        contexts.remove(classLoader);
    }

    private String[] getContextLocations(Class<P> pluginClass) {
        if(contextLocations == null || contextLocations.length == 0) {
            return  new String[] {
                    "META-INF/services/" +pluginClass.getName() +"/spring.xml",
                    "META-INF/services/" +pluginClass.getSimpleName() +"/spring.xml"
            };
        } else {
            return contextLocations;
        }
    }

    private  ApplicationContext initApplicationContext(final ServiceLocator services, URL url, ClassLoader loader) {
        final ConfigurableApplicationContext applicationContext = createApplicationContext(url, loader);



        // Make service beans available in the context
        applicationContext.addBeanFactoryPostProcessor(new BeanFactoryPostProcessor() {
            public void postProcessBeanFactory(ConfigurableListableBeanFactory beanFactory) throws BeansException {
                for(ServiceKey serviceKey : services.getServiceKeys()) {
                    @SuppressWarnings("unchecked")
                    Object bean = services.lookupService(serviceKey);


                    if(serviceKey.getServiceName() == null) {
                        // Nameless services are easy to register
                        beanFactory.registerResolvableDependency(serviceKey.getServiceType(), bean);
                    } else {
                        // Named services need to be defined first
                        BeanDefinitionRegistry registry = (BeanDefinitionRegistry) beanFactory;

                        RootBeanDefinition definition = new RootBeanDefinition(bean.getClass());

                        registry.registerBeanDefinition(serviceKey.getServiceName(), definition);

                        beanFactory.registerSingleton(serviceKey.getServiceName(), bean);
                    }
                }
            }
        });

        // Autowire programatically so plugin writers won't have to declare that
        applicationContext.addBeanFactoryPostProcessor(new BeanFactoryPostProcessor() {
            public void postProcessBeanFactory(ConfigurableListableBeanFactory beanFactory) throws BeansException {

                final AutowiredAnnotationBeanPostProcessor autowireProcessor = new AutowiredAnnotationBeanPostProcessor();
                autowireProcessor.setBeanFactory(beanFactory);
                beanFactory.addBeanPostProcessor(autowireProcessor);

            }
        });

        // Add any user-defined BeanFactoryPostProcessors
        addUserDefinedBeanFactoryPostProcessors(applicationContext);

        // Give subclasses a change to customize the context
        customizeContext(applicationContext);

        // Start the context
        applicationContext.refresh();

        return applicationContext;
    }

    private void addUserDefinedBeanFactoryPostProcessors(ConfigurableApplicationContext applicationContext) {
        if(this.beanFactoryPostProcessors != null) {
            for(BeanFactoryPostProcessor postProcessor : this.beanFactoryPostProcessors) {
                applicationContext.addBeanFactoryPostProcessor(postProcessor);
            }
        }
    }

    /**
     * The default implementation creates a ClassPathXmlApplicationContext. Subclasses may override this method
     * to create other context types such as an XmlWebApplicationContext.
     * @param url
     * @param loader
     * @return
     */
    protected ConfigurableApplicationContext createApplicationContext(URL url, ClassLoader loader) {
        final ClassPathXmlApplicationContext context = new ClassPathXmlApplicationContext();
        context.setClassLoader(loader);
        context.setConfigLocation(url.toExternalForm());
        return context;
    }

    /**
     * The default implementation is empty. Sublasses can override this method to customisze the context.
     * @param context
     */
    protected void customizeContext(ConfigurableApplicationContext context) {

    }

    /**
     * Set a list of {@link org.springframework.beans.factory.config.BeanFactoryPostProcessor}s that will be
     * added to the ApplicationContext for all plugins.
     * @param beanFactoryPostProcessors
     */
    public void setBeanFactoryPostProcessors(List<BeanFactoryPostProcessor> beanFactoryPostProcessors) {
        this.beanFactoryPostProcessors = beanFactoryPostProcessors;
    }
}
